import pandas as pd
import pkg_resources


def load_tourism() -> pd.DataFrame:
    """Load tourism Dataset

    :return:
    """
    stream = pkg_resources.resource_stream(__name__, "data/Tourism.csv")

    return pd.read_csv(stream)


import numpy as np
from typing import Union
from .hierarchy import Hierarchy


def _lamb_estimate(x: np.ndarray) -> float:
    """estimate :math`\\lambda` used in :ref:`shrinkage` estimator of mint method

    :param x: in-sample 1-step-ahead forecast error
    :return: :math`\\lambda`
    """
    T = x.shape[0]
    covm = x.T.dot(x)/T
    xs = x / np.sqrt(np.diag(covm))
    corm = xs.T.dot(xs)/T
    np.fill_diagonal(corm, 0)
    d = np.sum(np.square(corm))
    xs2 = np.square(xs)
    v = 1/(T*(T-1))*(xs2.T.dot(xs2) - 1/T*np.square(xs.T.dot(xs)))
    np.fill_diagonal(v, 0)
    lamb = np.max(np.min([np.sum(v)/d, 1]), 0)
    return lamb


def wls(hierarchy: Hierarchy,
        error: np.ndarray = None,
        method: str = "ols",
        weighting: Union[str, np.ndarray, None] = None,
        constraint_level: int = -1) -> np.ndarray:
    """function for forecast reconciliation.

    :param hierarchy: history time series.
    :param error: insample error.
    :param method: method used for forecast reconciliation, e.g. ols, wls, mint.
    :param weighting:
        method for calculating weight matrix used in reconciliation method, e.g. covariance matrix in mint
        or wls.
    :param constraint: If some levels are constrained to be unchangeable when reconciling base forecasts.
    :param constraint_level: Which level is constrained to be unchangeable when reconciling base forecasts.
    :return: reconciled forecasts
    """
    S = hierarchy.s_mat
    n = S.shape[0]
    m = S.shape[1]
    if method == "mint":
        T = error.shape[1]
        W = error.dot(error.T) / T
        if weighting == "variance":
            weight_matrix = np.diag(np.diagonal(W))
        elif weighting == "sample":
            weight_matrix = W
            if not np.all(np.linalg.eigvals(weight_matrix) > 0):
                raise ValueError("Sample method needs covariance matrix to be positive definite")
        elif weighting == "shrinkage":
            lamb = _lamb_estimate(error)
            weight_matrix = lamb * np.diag(np.diag(W)) + (1 - lamb) * W
        else:
            raise NotImplementedError("this min_trace method has not been implemented")

    elif method == "ols":
        weight_matrix = np.identity(n)
    else:
        if isinstance(weighting, np.ndarray):
            weight_matrix = np.linalg.inv(weighting)
        elif weighting == "structural":
            weight_matrix = np.diag(S.dot(np.array([1]*m)))
        else:
            raise ValueError("this wls weights is not supported now.")
    G = compute_g_mat(hierarchy, weight_matrix, constraint_level)
    return G


def _construct_u_mat(hierarchy: Hierarchy, constraint_level=-1):
    """construct U' mat used in solution

    :param s_mat:
    :param constraint_level:
    :return:
    """
    s_mat = hierarchy.s_mat
    n, m = s_mat.shape
    u1 = np.identity(n - m)
    u2 = 0-s_mat[:(n-m), :].astype('int32')
    u_mat = np.concatenate([u1, u2], axis=1)
    if constraint_level < 0:
        return u_mat.T
    u_up = np.identity(n)[hierarchy.node_level == constraint_level]
    return np.concatenate([u_up, u_mat], axis=0).T


def compute_g_mat(hierarchy: Hierarchy, weight_matrix, constraint_level=-1):
    """compute G matrix given weight_matrix

    :param hierarchy:
    :param weight_matrix:
    :param constraint_level:
    :return:
    """
    n, m = hierarchy.s_mat.shape
    u = _construct_u_mat(hierarchy, constraint_level=constraint_level)
    c = np.concatenate([np.zeros([m, n-m]), np.identity(m)], axis=1)
    a = np.zeros([n - m, n])
    if constraint_level >= 0:
        a = np.concatenate([np.identity(n)[hierarchy.node_level == constraint_level], a])
    return c - c.dot(weight_matrix).dot(u).dot(np.linalg.inv((u.T.dot(weight_matrix).dot(u)))).dot(u.T-a)


__all__ = [
    "wls"
]





import numpy as np


def mase(y_true, y_pred, history, m):
    q = np.abs(history[m:] - history[:(len(history) - m)]).mean()
    e = np.abs(y_true - y_pred).mean()
    return e/q


def mse(y_true, y_pred, history, m):
    return np.mean(np.square(y_true-y_pred))


def mape(y_true, y_pred, history, m):
    return np.mean(np.abs((y_true-y_pred)/y_true))


def rmse(y_true, y_pred, history, m):
    return np.sqrt(mse(y_true, y_pred, history, m))


def mae(y_true, y_pred, history, m):
    return np.mean(np.abs(y_true - y_pred))


def smape(y_true, y_pred, history, m):
    return 2 * np.mean((y_true-y_pred)/(np.abs(y_true)+np.abs(y_pred)))


def rmsse(y_true, y_pred, history, m):
    q = np.square((history[m:] - history[:(len(history) - m)]).mean())
    e = np.square((y_true-y_pred).mean())
    return np.sqrt(e/q)




import numpy as np
import rpy2.robjects as robjects
from rpy2.robjects.packages import importr
from rpy2.robjects.vectors import FloatVector
forecast = importr("forecast")
ts = robjects.r["ts"]


class BaseForecaster:
    """Base Class for forecaster

    """
    def __init__(self):
        self.fitted = None

    def fit(self, hist, **kwargs):
        raise NotImplementedError

    def forecast(self, h: int, **kwargs) -> np.ndarray:
        raise NotImplementedError


class AutoArimaForecaster(BaseForecaster):
    """auto.arima forecaster.

    """

    def __init__(self, m: int = 1):
        super().__init__()
        self.m = m
        self.hist = None
        self.model = None

    def fit(self, hist: np.ndarray, x_reg=None, **kwargs):
        self.hist = ts(FloatVector(hist), frequency=self.m)
        arima = forecast.auto_arima
        self.model = arima(self.hist, xreg=x_reg if x_reg is not None else robjects.r('NULL'), **kwargs)
        self.fitted = self.model.rx2["fitted"]
        return self

    def forecast(self, h: int, x_reg=None, **kwargs) -> np.ndarray:
        return np.array(forecast.forecast(self.model, h=h, xreg=x_reg if x_reg is not None else robjects.r('NULL'),
                                          **kwargs).rx2["mean"])


class EtsForecaster(BaseForecaster):
    """Ets forecaster.

    """

    def __init__(self, m: int = 1):
        super().__init__()
        self.m = m
        self.hist = None
        self.model = None

    def fit(self, hist: np.ndarray, x_reg=None, **kwargs):
        self.hist = ts(FloatVector(hist), frequency=self.m)
        ets = forecast.ets
        self.model = ets(self.hist, **kwargs)
        self.fitted = self.model.rx2["fitted"]
        return self

    def forecast(self, h: int, **kwargs) -> np.ndarray:
        return np.array(forecast.forecast(self.model, h=h).rx2["mean"])



import numpy as np
import pandas as pd
from copy import copy
from typing import Union, List
from . import accuracy


class Hierarchy:
    """Class for hierarchy structure.

    **Attributes**

        .. py:attribute:: s_mat

            summing matrix.
        .. py:attribute:: period

            frequency of time series.
        .. py:attribute:: node_name

            name of each node.
        .. py:attribute:: level_n

            number of levels.
        .. py:attribute:: node_level

            level of each node.
    """

    def __init__(self, s_mat, node_level, names, period):
        self.s_mat = s_mat.astype('int8')
        """ summing matrix. """
        self.node_level = np.array(node_level)
        self.level_n = max(node_level) + 1
        self.node_name = np.array(names)
        self.period = period

    @classmethod
    def from_node_list(cls, nodes, period: int = 1) -> "Hierarchy":
        """Constructs hierarchy from a list of lists, each sublist contains number of children nodes of all nodes in a
        hierarchical level. More specifically, the first sublist contains the number of children nodes of all nodes in
        Level0, e.g. the root node.(sum of all bottom level time series), and so on.

        **Examples**

            >>> from pyhts.hierarchy import Hierarchy
            >>> nodes = [[2], [2, 2]]
            >>> hierarchy = Hierarchy.from_node_list(nodes, 12)
            >>> hierarchy.s_mat
            array([[1, 1, 1, 1],
                   [1, 1, 0, 0],
                   [0, 0, 1, 1],
                   [1, 0, 0, 0],
                   [0, 1, 0, 0],
                   [0, 0, 1, 0],
                   [0, 0, 0, 1]], dtype=int8)

        :code:`[[2], [2, 2]]` defines a hierarchy that in which the root node have two sub-nodes, the fist
        sub-node contains two sub-nodes and the second sub-node also contains two sub-nodes. There are 4 bottom time
        series in total.

        :param nodes: list of children nodes number lists.
        :param period: frequency of the time series, 1 means non-seasonality data, 12 means monthly data.
        :return:
        """
        nodes = copy(nodes)
        n = sum(nodes[-1])
        m = sum(map(sum, nodes)) + 1
        node_level = n * [len(nodes)]
        nodes.append([1] * n)
        s = np.zeros([m, n])
        c_row_start = m - n
        s[c_row_start:, :] = np.identity(n)
        bts_count = nodes[-1]
        for level_idx in range(len(nodes) - 2, -1, -1):
            c_cum = 0
            c_x = 0
            level = nodes[level_idx]
            c_row_start = c_row_start - len(level)
            new_bts_count = []
            c_row = c_row_start
            for node_idx in range(len(nodes[level_idx])):
                n_x = c_x + level[node_idx]
                new_bts_count.append(sum(bts_count[c_x:n_x]))
                n_cum = c_cum + new_bts_count[-1]
                s[c_row, c_cum:n_cum] = 1
                c_cum = n_cum
                c_row += 1
                c_x = n_x
                node_level.insert(0, level_idx)
            bts_count = new_bts_count
        return cls(s, np.array(node_level), node_level, period)

    @classmethod
    def from_names(cls, names: List[str], chars: List[int], period: int = 1) -> "Hierarchy":
        """Construct Hierarchy from column names of bottom time series. The name of bottom series should be consist of
        several parts. Each part points to a specific level of the hierarchy and it should have fixed length that used to
        split and recognize hierarchy.

        For example :code:`AA`. The first :code:`A` represents :code:`A` series in level1,
        the second :code:`A` represents the :code:`A` series in level 2 and its parent node is :code:`A` series in level 1. This method
        will add a :code:`Total` level.

        **Examples**

            >>> from pyhts.hierarchy import Hierarchy
            >>> names = ['AA', 'AB', 'BA', 'BB']
            >>> hierarchy = Hierarchy.from_names(names, chars=[1, 1], period=12)
            >>> hierarchy.s_mat
            array([[1, 1, 1, 1],
                   [1, 1, 0, 0],
                   [0, 0, 1, 1],
                   [1, 0, 0, 0],
                   [0, 1, 0, 0],
                   [0, 0, 1, 0],
                   [0, 0, 0, 1]], dtype=int8)

        The example above define a hierarchy same as the hierarchy in example of
        :meth:`~pyhts.hierarchy.Hierarchy.from_node_list()`.

        :param names: columns of all **bottom** level time series.
        :param chars: character length of each part.
        :param period: frequency of the time series, 1 means non-seasonality data, 12 means monthly data.
        :return:
        """
        df = pd.DataFrame()
        df['bottom'] = names
        df['top'] = 'Total'
        total_index = 0
        names = ['Total']
        node_level = [0]
        index = 0
        for index in range(len(chars)-1):
            df[index+1] = df['bottom'].apply(lambda x: x[:total_index+chars[index]])
            total_index += chars[index]
            names += list(df[index+1].unique())
            node_level += [index+1] * len(df[index+1].unique())
        cols = ['top', *[i+1 for i in range(len(chars)-1)], 'bottom']
        df = df[cols]
        s_mat = pd.get_dummies(df).values.T
        names += list(df['bottom'])
        node_level += [index+2] * len(df['bottom'])
        return cls(s_mat, np.array(node_level), names, period)

    @classmethod
    def from_balance_group(cls, group_list: List[List[str]], period=1) -> "Hierarchy":
        """Construct group hierarchical structure.

        **Examples**

            >>> from pyhts.hierarchy import Hierarchy
            >>> groups = [['A', 'B'], ['Item1', 'Item2']]
            >>> hierarchy = Hierarchy.from_balance_group(groups)
            >>> hierarchy.node_name
            array(['Total', 'A', 'B', 'Item1', 'Item2', 'A_Item1', 'A_Item2',
                   'B_Item1', 'B_Item2'], dtype='<U7')
            >>> hierarchy.s_mat
            array([[1, 1, 1, 1],
                   [1, 1, 0, 0],
                   [0, 0, 1, 1],
                   [1, 0, 1, 0],
                   [0, 1, 0, 1],
                   [1, 0, 0, 0],
                   [0, 1, 0, 0],
                   [0, 0, 1, 0],
                   [0, 0, 0, 1]], dtype=int8)


        :param group_list: Two group types are supported for now.
        :param period: frequency of the time series, 1 means non-seasonality data, 12 means monthly data.
        :return:
        """
        from itertools import product
        df = pd.DataFrame(product(*group_list))
        cols = ['top', *[i for i in range(len(group_list))], 'bottom']
        df['bottom'] = df.apply(lambda x: '_'.join(x), axis=1)
        df['top'] = 'total'
        df = df[cols]
        dummy_df = pd.get_dummies(df)
        s_mat = dummy_df.values.T
        node_level = [0]
        names = ['Total']
        i = 0
        for i in range(len(group_list)):
            node_level += [i+1]*len(group_list[i])
            names += group_list[i]
        node_level += [i+2] * df.shape[0]
        names += list(df['bottom'])
        return cls(s_mat, np.array(node_level), names, period)

    @classmethod
    def from_long(cls, df: pd.DataFrame, keys: List[str], period=1) -> "Hierarchy":
        """Construct hierarchy from long data table that each row represents a bottom level time series. This method is
        suitable for complex hierarchical structure.

        **Examples**

            >>> from pyhts.hierarchy import Hierarchy
            >>> df = pd.DataFrame({"City": ["A", "A", "B", "B"], "Store": ["Store1", "Store2", "Store3", "Store4"]})
            >>> hierarchy = Hierarchy.from_long(df, ["City", "Store"])
            >>> hierarchy.node_name
            array(['Total', 'A', 'B', 'Store1', 'Store2', 'Store3', 'Store4'],
                  dtype='<U8')
            >>> hierarchy.s_mat
            array([[1, 1, 1, 1],
                   [1, 1, 0, 0],
                   [0, 0, 1, 1],
                   [1, 0, 0, 0],
                   [0, 1, 0, 0],
                   [0, 0, 1, 0],
                   [0, 0, 0, 1]], dtype=int8)
            >>> df = pd.DataFrame({"City": ["A", "A", "B", "B"], "Category": ["C1", "C2", "C1", "C2"]})
            >>> hierarchy = Hierarchy.from_long(df, ["City", "Category"])
            >>> hierarchy.s_mat
            array([[1, 1, 1, 1],
                   [1, 1, 0, 0],
                   [0, 0, 1, 1],
                   [1, 0, 1, 0],
                   [0, 1, 0, 1],
                   [1, 0, 0, 0],
                   [0, 1, 0, 0],
                   [0, 0, 1, 0],
                   [0, 0, 0, 1]], dtype=int8)

        :param df: DataFrame contains keys for determining hierarchical structural.
        :param keys: column names that each column represents a level.
        :param period: frequency of the time series, 1 means non-seasonality data, 12 means monthly data.
        :return:
        """
        new_df = copy(df[keys]).astype('string')
        new_df['Total'] = 'Total'
        node_level = [0] + [1] * len(new_df[keys[0]].unique())
        names = ['Total'] + list(new_df[keys[0]].unique())
        n = new_df.shape[0]
        for index in range(1, len(keys)):
            names += list(new_df[keys[index]].unique())
            node_level += [index+1] * len(new_df[keys[index]].unique())

        if len(new_df[keys[-1]].unique()) != new_df.shape[0]:
            new_df['bottom'] = new_df[keys].apply(lambda x: '-'.join(x), axis=1)
            names += list(new_df['bottom'])
            node_level += [len(keys)+1] * n
            s_mat = pd.get_dummies(new_df[['Total'] + keys + ['bottom']], columns=['Total'] + keys + ['bottom']).values.T
        else:
            s_mat = pd.get_dummies(new_df[['Total'] + keys], columns=["Total"] + keys).values.T
        return cls(s_mat, np.array(node_level), names, period=period)

    def aggregate_ts(self, bts: np.ndarray, levels: Union[int, List[int]] = None) -> np.ndarray:
        """aggregate bottom-levels time series.

        :param levels: which levels you want.
        :param bts: bottom time series
        :return: upper-level time series.
        """
        if isinstance(levels, int):
            s = self.s_mat[np.where(self.node_level == levels)]
            return s.dot(bts.T).T
        if isinstance(levels, list):
            s = self.s_mat[np.isin(self.node_level, levels)]
            return s.dot(bts.T).T
        return self.s_mat.dot(bts.T).T

    def check_hierarchy(self, *hts):
        for ts in hts:
            if ts.shape[1] == self.s_mat.shape[1]:
                return True
            else:
                return False

    def accuracy_base(self, real, pred, hist=None,
                      levels: Union[int, None, List] = None,
                      measure: List[str] = None) -> pd.DataFrame:
        """calculate forecast accuracy of base forecast.

        :param real: real future observations, array-like of shape (forecast_horizon, m)
        :param pred: forecast values, array-like of shape (forecast_horizon, n)
        :param levels: which level, None means all levels.
        :param hist: history time series.
        :param measure: list of measures, e.g. ['mase'], ['mse', 'mase'].
        :return: forecast accuracy of base forecasts.
        """
        assert self.check_hierarchy(real), f"true observations should be of shape (h, m)"
        assert real.shape[0] == pred.shape[0], \
            f" {real.shape} true observations have different length with {real.shape} forecasts"
        agg_true = self.aggregate_ts(real, levels=levels)
        if measure is None:
            measure = ['mase', 'mape', 'rmse']
        if 'mase' in measure or 'smape' in measure or 'rmsse' in measure:
            assert hist is not None
            hist = self.aggregate_ts(hist, levels=levels)
        if hist is None:
            hist = [None] * self.s_mat.shape[0]
        accs = pd.DataFrame()
        for me in measure:
            try:
                accs[me] = np.array(list(map(lambda x: getattr(accuracy, me)(*x),
                                             zip(agg_true.T, pred.T, hist.T, [self.period] * self.s_mat.shape[0]))))
            except AttributeError:
                print(f'Forecasting measure {me} is not supported!')
        accs.index = self.node_name[np.isin(self.node_level, levels)]
        return accs

    def accuracy(self, real, pred, hist=None,
                 levels: Union[int, None, List] = None,
                 measure: List[str] = None):
        """calculate forecast accuracy.

        :param real: real future observations, array-like of shape (forecast_horizon, m)
        :param pred: forecast values, array-like of shape (forecast_horizon, m)
        :param levels: which level, None means all levels.
        :param hist: history time series.
        :param measure: list of measures, e.g. ['mase'], ['mse', 'mase'].
        :return: forecast accuracy of reconciled forecasts.
        """
        assert self.check_hierarchy(real), f"true observations should be of shape (h, m)"
        assert self.check_hierarchy(pred), f"forecast values should be of shape (h, m)"
        assert real.shape[0] == pred.shape[0], \
            f" {real.shape} true observations have different length with {real.shape} forecasts"
        if measure is None:
            measure = ['mase', 'mape', 'rmse']
        if 'mase' in measure or 'smape' in measure or 'rmsse' in measure:
            assert hist is not None
            hist = self.aggregate_ts(hist, levels=levels).T
        agg_true = self.aggregate_ts(real, levels=levels)
        agg_pred = self.aggregate_ts(pred, levels=levels)
        accs = pd.DataFrame()
        for me in measure:
            try:
                accs[me] = np.array(list(map(lambda x: getattr(accuracy, me)(*x),
                                             zip(agg_true.T, agg_pred.T, hist, [self.period] * self.s_mat.shape[0]))))
            except AttributeError:
                print('this forecasting measure is not supported!')
        accs.index = self.node_name[np.isin(self.node_level, levels)]
        return accs



from __future__ import annotations
import numpy as np
from pyhts import accuracy
from pandas import DataFrame
from copy import copy

from typing import List, Union, Optional
from rpy2.robjects.packages import importr
from scipy.sparse import csr_matrix
from pyhts.forecaster import BaseForecaster, AutoArimaForecaster, EtsForecaster

forecast = importr("forecast")


def _nodes2constraints(nodes: List[List[int]]) -> [csr_matrix, np.ndarray]:
    """construct constraints from nodes

    :param nodes: nodes that demonstrate hierarchy
    :return: constraints
    """
    nodes = copy(nodes)
    n = sum(nodes[-1])
    m = sum(map(sum, nodes)) + 1
    node_level = n * [len(nodes)]
    nodes.append([1]*n)
    s = np.zeros([m, n])
    c_row_start = m - n
    s[c_row_start:, :] = np.identity(n)
    bts_count = nodes[-1]
    for level_idx in range(len(nodes)-2, -1, -1):

        c_cum = 0
        c_x = 0
        level = nodes[level_idx]
        c_row_start = c_row_start - len(level)
        new_bts_count = []
        c_row = c_row_start
        for node_idx in range(len(nodes[level_idx])):
            n_x = c_x+level[node_idx]
            new_bts_count.append(sum(bts_count[c_x:n_x]))
            n_cum = c_cum + new_bts_count[-1]
            s[c_row, c_cum:n_cum] = 1
            c_cum = n_cum
            c_row += 1
            c_x = n_x
            node_level.insert(0, level_idx)
        bts_count = new_bts_count
    return csr_matrix(s), np.array(node_level)


def _constraints_from_chars(names: List, chars: List) -> [csr_matrix, np.ndarray]:
    """construct constraints from columns and characters.

    :param names: names that represents location of series in hierarchy.
    :param chars: characters.
    :return: constraints
    """
    total_pos = 0
    import pandas as pd
    df = pd.DataFrame()
    df["0"] = [1]*len(names)
    for index in range(len(chars)):
        a = list(map(lambda x: x[:(total_pos+chars[index])], names))
        df[f"{index+1}"] = a
        total_pos += chars[index]

    constraints = pd.get_dummies(df).values.T

    return csr_matrix(constraints), np.array(list(map(lambda x: int(x.split("_")[0]), pd.get_dummies(df).columns)))


class Hts:
    """
    Class for hierarchical time series, can be constructed from  cross-sectional hierarchical time series
    like hierarchical or grouped time series and temporal hierarchies :ref:`[1]<references>`.

    """
    def __init__(self, constraints: Union[csr_matrix, np.ndarray], bts: np.ndarray, node_level: np.ndarray, m: int = 1):
        """initialize a Hts object according to attributes directly. Do not use it unless you are familiar with meanings
        of parameters.

        :param constraints: constraints, also called "summing matrix" in some context.
        :param bts: bottom level time series.
        :param node_level: which level every nodes belong to.
        :param m: frequency of time series data.
        """
        if isinstance(constraints, np.ndarray):
            constraints = csr_matrix(constraints)
        self.constraints = constraints
        self.bts = bts
        self.node_level = node_level
        self.m = m
        self.base_forecast = None
        self.keep_fitted = False

    def aggregate_ts(self, levels: Union[int, List[int]] = None) -> np.ndarray:
        """aggregate bottom-levels time series.

        :param levels: which levels you want.
        :return: upper-level time series.
        """
        if isinstance(levels, int):
            s = self.constraints[np.where(self.node_level == levels)]
            return s.dot(self.bts.T).T
        if isinstance(levels, list):
            s = self.constraints[np.isin(self.node_level, levels)]
            return s.dot(self.bts.T).T
        return self.constraints.dot(self.bts.T).T

    def forecast(self,
                 h: int,
                 base_method: Union[str, None] = "arima",
                 base_forecaster: Union[List, None, BaseForecaster] = None,
                 hf_method: str = "comb",
                 comb_method: str = "ols",
                 weights: Optional[np.ndarray, str] = None,
                 parallel: bool = False,
                 constraint: bool = False,
                 constrain_level: int = 0,
                 keep_base_forecast: bool = True
                 ) -> Hts:
        """forecast Hts using specific reconciliation methods and base methods.

        :param h: forecasting horizon.
        :param base_method:
            method for generate base forecast, arima, ets, or custom function.
            arima and ets are implemented using `forecast` package in R.
            If custom forecast function is specified and using "mint" method, the function
            should return the in-sample forecasts for estimating the covariance matrix, e.g. the
            base forecast returned should be :math:`(T+h) \\times n`.
        :param base_forecaster:
            list for base forecasters of each level. If you want different base forecast methods
            for different levels, just pass a list of base forecasters, see :doc:forecaster.
        :param hf_method: method for hierarchical forecasting, "comb", "bu", "td", "mo"
        :param comb_method:
            "ols", "wls", "mint"(e.g Minimum Trace :ref:`[2]<mint>` ), weights method used for "comb"(e.g. optimal
            combination) reconciliation method, if you choose "wls", you should specify `weights`, or the result is same
            as ols. If you choose "mint", you should specify `variance` parameter.
        :param weights:
            weighting matrix used for `wls` combination method and variance used for `mint` combination method,
            if `wls`, can be "structural" or custom_matrix, this custom matrix should be
            :math:`n\\times n` symmetric matrix.
            If `mint`, can be "sample", "variance" or "shrinkage", please refer to :doc:`/tutorials/reconciliation`.
        :param parallel: If parallel, not supported for now.
        :param constraint: If some levels are constrained to be unchangeable when reconciling base forecasts.
        :param constrain_level: Which level is constrained to be unchangeable when reconciling base forecasts.
        :param keep_base_forecast:
            if keep keep_base_forecast, if True, attribute `Hts.base_forecast` is set. if false, it will be None
        :return: Hts: reconciled forecast.
        """
        import pyhts.reconciliation as fr
        if hf_method == "comb":
            # generate base forecast
            if comb_method == "mint":
                self.keep_fitted = True
            if base_method == "arima":
                base_forecaster = [AutoArimaForecaster(self.m)]*int(max(self.node_level)+1)
            elif base_method == "ets":
                base_forecaster = [EtsForecaster(self.m)]*int(max(self.node_level)+1)
            elif base_method is not None:
                raise ValueError("this base forecast method is not supported now.")
            else:
                if base_forecaster is None:
                    raise ValueError("You should either give base_method of base_forecaster")
                elif not isinstance(base_forecaster, List):
                    base_forecaster = [base_forecaster] * int(max(self.node_level)+1)
                else:
                    assert len(base_forecaster) == int(max(self.node_level)+1)
            base_forecast = self.generate_base_forecast(base_forecaster, h, self.keep_fitted)
            # reconcile base forecasts
            if comb_method == "ols":
                reconciled_y = fr.wls(self, base_forecast, method="ols")
            elif comb_method == "wls":
                if weights == "structural":
                    reconciled_y = fr.wls(self, base_forecast, method="wls", weighting="structural",
                                          constraint=constraint, constraint_level=constrain_level)
                elif isinstance(weights, np.ndarray):
                    reconciled_y = fr.wls(self, base_forecast, method="wls", weighting=weights,
                                          constraint=constraint, constraint_level=constrain_level)
                else:
                    raise ValueError("weights for wls method is not supported")
            elif comb_method == "mint":
                reconciled_y = fr.wls(self, base_forecast, method="mint", weighting=weights,
                                      constraint=constraint, constraint_level=constrain_level)
            else:
                raise ValueError("this comination method is not supported")
        else:
            raise NotImplementedError("this method is not implemented")
        if keep_base_forecast:
            self.base_forecast = base_forecast
        return reconciled_y

    def generate_base_forecast(self, method: List[BaseForecaster], h: int = 1, keep_fitted: bool = False) -> np.ndarray:
        """generate base forecasts by `forecast` in R with rpy2.

        :param method: base forecast method.
        :param h: forecasting horizons.
        :param keep_fitted: if keep in-sample fitted value, useful when mint method is specified.
        :return: base forecast
        """
        k = int(max(self.node_level)+1)
        length = self.bts.shape[0]
        n = len(self.node_level)
        if keep_fitted:
            f_casts = np.zeros([h + length, n])
        else:
            f_casts = np.zeros([h, n])
        j = 0
        for i in range(k):
            aggts = self.aggregate_ts(levels=i)
            forecaster = method[i]
            for ts in range(aggts.shape[1]):
                f_casts[:, j] = forecaster.forecast(hist=aggts[:, ts], h=h, keep_fitted=keep_fitted)
                j += 1
        return f_casts

    @classmethod
    def from_temporal_hierarchy(cls, ts: np.ndarray, m: int, aggregate_lens: List[int]):
        """Construct Hts from temporal hierarchy.

        :param ts: a time series.
        :param m: frequency
        :param aggregate_lens:
            length of bottom level time series used for aggregation. For example, for monthly data,
            3 means quarterly data, 6 mean half-annual data, 12 means annual time series.  The time
            series should include 1 at least.
        :return: Hts object.
        """
        aggregate_lens.sort()

        smatrix = np.zeros([sum([int(m / k) for k in aggregate_lens]), m])
        node_level = []
        if len(ts) % m != 0:
            Warning("length of history time series is not multiple of m, some observations at very beginning "
                    "would be cut out.")
            ts = ts[len(ts) % m:]

        if aggregate_lens[0] != 1:
            raise ValueError("You'd better always include 1 in the aggregate_lens")

        aggregate_lens.reverse()
        index = 0
        for i in range(len(aggregate_lens)):
            k = aggregate_lens[i]
            if (m % k) != 0:
                raise ValueError("aggregate length should be factor of m")
            mk = m // k
            for j in range(mk):
                smatrix[index, j * k:(j + 1) * k] = 1
                index += 1
                node_level.append(i)
        constraints = csr_matrix(smatrix)
        bts = ts.reshape([len(ts) // m, m])
        node_level = np.array(node_level)
        hts = Hts(constraints, bts, node_level, m)
        return hts

    @classmethod
    def from_hts(cls, bts: Union[np.ndarray, DataFrame],
                 m: int,
                 characters: Optional[List[int]] = None,
                 nodes: Optional[List[List]] = None) -> Hts:
        """Construct hts from cross-sectional hierarchical time series.

        :param bts: :math:`T\\times m` bottom level series.
        :param m: frequency of time series
        :param characters:
            length of characters of each level,
            for example, "ABC" 'A' represents 'A' series of  first level,  'B' represents 'B' series under node 'A',
            'C' represents 'C' series under node 'AB'. So the value of parameter is [1,1,1]
        :param nodes: List of list to demonstrate the hierarchy, see details.
        :return: Hts object.
        """
        if isinstance(bts, DataFrame):
            names = bts.columns
            bts = bts.values
        elif isinstance(bts, np.ndarray):
            bts = bts
            names = None
        else:
            raise TypeError("bts must be numpy.ndarray or pandas.DataFrame")
        if nodes is not None:
            constraints, node_level = _nodes2constraints(nodes)
        elif characters is not None:
            constraints, node_level = _constraints_from_chars(list(names), characters)
        else:
            constraints, node_level = _nodes2constraints([[bts.shape[1]]])
        hts = cls(constraints, bts, node_level, m)
        return hts

    def accuracy(self, y_true: Hts, y_pred: Hts, levels: Union[int, None, List] = None, measure: List[str] = None) -> Union[float, DataFrame]:
        """calculate forecast accuracy, mase is supported only for now.

        :param y_true: real observations.
        :param y_pred: forecasts.
        :param levels: which level.
        :param measure:
            mase, mse is supported for now. e.g. ['mase'], ['mse', 'mase'].
            if None, mase is calculated.
        :return: forecast accuracy.
        """
        if measure is None:
            measure = ['mase']
        agg_ts = self.aggregate_ts(levels=levels)
        agg_true = y_true.aggregate_ts(levels=levels)
        agg_pred = y_pred.aggregate_ts(levels=levels)
        accs = DataFrame()
        for me in measure:
            try:
                accs[me] = np.array(list(map(lambda x: getattr(accuracy, me)(*x),
                                             zip(agg_ts.T, agg_true.T, agg_pred.T, [self.m] * agg_ts.shape[1]))))
            except AttributeError:
                print('this forecasting measure is not supported!')
        return accs

    def accuracy_base(self, y_true: Hts, levels: Union[int, None, List] = None, measure: List[str] = None) -> DataFrame:
        """calculate forecast accuracy of base forecast

        :param y_true: real observations.
        :param levels: which level.
        :param measure:
            mase, mse is supported for now. e.g. ['mase'], ['mse', 'mase'].
            if None, mase is calculated.
        :return: forecast accuracy of base forecasts.
        """
        agg_ts = self.aggregate_ts(levels=levels)
        agg_true = y_true.aggregate_ts(levels=levels)
        horizon = agg_true.shape[0]
        agg_pred = self.base_forecast[-horizon:, :]
        if measure is None:
            measure = ['mase']
        accs = DataFrame()
        for me in measure:
            try:
                accs[me] = np.array(list(map(lambda x: getattr(accuracy, me)(*x),
                                             zip(agg_ts.T, agg_true.T, agg_pred.T, [self.m] * agg_ts.shape[1]))))
            except AttributeError:
                print(f'Forecasting measure {me} is not supported!')
        return accs


if __name__ == '__main__':
    print(_nodes2constraints([[2], [2, 2]]))



import numpy as np
from typing import List, Union, Optional
from .forecaster import BaseForecaster, EtsForecaster, AutoArimaForecaster
from . import reconciliation as fr
from .hierarchy import Hierarchy
import pandas as pd


class HFModel:
    """Model for hierarchical forecasting.

    """

    def __init__(self,
                 hierarchy: Hierarchy,
                 base_forecasters: Union[List[BaseForecaster], str],
                 hf_method: str = 'comb',
                 comb_method: str = 'ols',
                 weights: Optional[Union[str, np.ndarray]] = None,
                 constrain_level: int = -1):
        """Define a Hierarchical Forecasting Model.

        :param hierarchy: hierarchical structure of data.
        :param base_forecasters: base forecasting method, custom forecasters or arima, ets, implemented by :code:`forecast` in R called by :code:`rpy2`
        :param hf_method: method for hierarchical forecasting, :code:`comb` is only supported for now, which represents optimal combination method.
        :param comb_method: method for forecast reconciliation, ols, wls or mint.
        :param weights: weighting matrix used in wls and mint, "structural" or custom symmetric matrix for wls. "shrinkage", "sample", "variance" for mint.
        :param constrain_level: -1 means no constraints.
        """
        self.hierarchy = hierarchy
        self.base_forecasters = base_forecasters
        self.hf_method = hf_method
        self.comb_method = comb_method
        self.period = self.hierarchy.period
        self.weights = weights
        self.constrain_level = constrain_level
        self.G = None

    def fit(self, ts: pd.DataFrame, x_reg: Optional[np.array]=None, **kwargs) -> "HFModel":
        """Fit base forecast model and calculate reconciliation matrix used for reconciliation.

        :param ts: DataFrame that each column contains a bottom time series.
        :param x_reg: explantory variables with shape (n, T, k), where n is number of time series, T is history length, k is dimension of explantory variables.
        :return: fitted HFModel
        """
        assert self.hierarchy.check_hierarchy(ts), "only bottom series are needed to fit the model."
        s_matrix = self.hierarchy.s_mat
        n, m = self.hierarchy.s_mat.shape
        try:
            ts = s_matrix.dot(ts[self.hierarchy.node_name[-m:]].T)
        except KeyError:
            ts = s_matrix.dot(ts)
        if isinstance(self.base_forecasters, str):
            if self.base_forecasters == 'arima':
                self.base_forecasters = [AutoArimaForecaster(m=self.period).fit(ts[i, :], x_reg=None if x_reg is None else x_reg[i])
                                         for i in range(n)]
            elif self.base_forecasters == 'ets':
                self.base_forecasters = [EtsForecaster(m=self.period).fit(ts[i, :], x_reg=None if x_reg is None else x_reg[i])
                                         for i in range(n)]
            else:
                raise ValueError("not supported base method")
        elif isinstance(self.base_forecasters, List):
            if len(self.base_forecasters) == 1:
                self.base_forecasters = [self.base_forecasters[0]().fit(ts[i, :], x_reg=None if x_reg is None else x_reg[i])
                                         for i in range(n)]
            if len(self.base_forecasters) == self.hierarchy.level_n:
                self.base_forecasters = [self.base_forecasters[self.hierarchy.node_level[i]]().fit(ts[i, :], x_reg=None if x_reg is None else x_reg[i])
                                         for i in range(n)]
            if len(self.base_forecasters) == n:
                self.base_forecasters = [self.base_forecasters[i]().fit(ts[i, :], x_reg=None if x_reg is None else x_reg[i])
                                         for i in range(n)]
        else:
            raise ValueError("not supported base method")

        if self.hf_method == "comb":
            if self.comb_method == "ols":
                self.G = fr.wls(self.hierarchy, error=None, method="ols",
                                constraint_level=self.constrain_level)
            elif self.comb_method == "wls":
                if self.weights == "structural":
                    self.G = fr.wls(self.hierarchy, error=None, method="wls", weighting="structural",
                                    constraint_level=self.constrain_level)
                elif isinstance(self.weights, np.ndarray):
                    self.G = fr.wls(self.hierarchy, error=None, method="wls", weighting=self.weights,
                                    constraint_level=self.constrain_level)
                else:
                    raise ValueError("weights for wls method is not supported")
            elif self.comb_method == "mint":
                sample_forecast = np.stack([model.fitted for model in self.base_forecasters], axis=0)
                sample_error = ts - sample_forecast
                self.G = fr.wls(self.hierarchy, sample_error, method="mint", weighting=self.weights,
                                constraint_level=self.constrain_level)
            else:
                raise ValueError("this comination method is not supported")
        else:
            raise NotImplementedError("this method is not implemented")
        return self

    def generate_base_forecast(self, horizon: int = 1, x_reg=None, **kwargs):
        forecasts = np.stack([self.base_forecasters[i].forecast(h=horizon, x_reg=None if x_reg is None else x_reg[i],
                                                                **kwargs) for i in range(len(self.base_forecasters))])
        return forecasts.T

    def predict(self, horizon: int = 1, x_reg=None, **kwargs) -> np.array:
        """generate horizon-step-ahead reconciled base forecasts of bottom level.

        :param horizon: forecast horizon
        :param x_reg: explantory variables with shape (n, h, k), where n is number of time series, h is forecast horizon, k is dimension of explantory variables.
        :param kwargs: other parameters passed to base forecasters
        :return: coherent forecasts of bottom level.
        """
        forecasts = self.generate_base_forecast(horizon=horizon, x_reg=x_reg, **kwargs)
        return self.G.dot(forecasts.T).T
